clear all;
close all;
clc;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Hamming Codes                            %
% - Can repair 1 Error or detect 2 Errors  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


P = 4;         % Number of Parity Bits (3 for hamming(7,4))
N = 2^P-1;     % Codeword length
B = N-P;       % Number of data bits
Repair = 1;    % Enable (1) / Disable (0) Repair 
NumErrors = 3; % How many Errors to simulate | Try these: (1,2,3)
disp(['Hamming CODE: [',num2str(N),',',num2str(B),']']);
disp(['Parity Bits: ',num2str(P),' Data bits: ',num2str(B),' Message Length: ',num2str(N)]);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Calculate Parity Positions %
LinearBase = eye(B);
Pows = ceil(log2(N));
Pows = 0:Pows-1;
ParityIndexes = 2.^(Pows);


%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Create Generator Matrix %
% - Add Linear Base
G = nan(B,N);
for i = 1:B
    jCounter = 1;
    for j = 1:N
        if( any(j == ParityIndexes))
        else
             G(i,j) = LinearBase(i,jCounter);
             jCounter = jCounter + 1;
        end
       
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ADD Parity Bits to generator Matrix %
for i = 1:B % ROWS
    JSkip = 1;
    Pbit = 1; 
    for j = 1:N % COLUMNS


        if( any(j == ParityIndexes))
            ToSkip = 2^(JSkip-1)-1;  % SKIP THIS NUMBER OF ELEMENETS
            BitSum = 0;              % SUM OF THE BITS USED TO FILL THE PARITY
            Start = 0;               % DO NOT READ THE FIRST VALUE
            for k = 1:N
                if(mod(k-1,2^(Pbit)) < 2^(Pbit-1))
                    if(Start)
                        BitSum = BitSum + G(i,k + ToSkip);
                    end
                    Start = 1;
                end
            end
            
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Fill Generator Matrix - Parity Position %
            if (mod(BitSum,2) == 0)
                G(i,j) = 0;
            else
                 G(i,j) = 1;
            end
            
            %%%%%%%%%%%%%%%%%%%%
            % Update Variables %
            JSkip = JSkip + 1;
            Pbit = Pbit +1;
        end
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Reassign G to Standard Shape %
Parities = G(:,ParityIndexes);
G = [eye(B),Parities];            % STANDARD SHAPE: [I|P]
H = [transpose(Parities),eye(P)]; % STANDARD SHAPE: [Pt|I]

disp('Generator Matrix: ');
disp(num2str(G));


%%%%%%%%%%%%%%%%%%%%%%%%%%
% Precalculate Syndromes %
S = zeros(N,P); % Matrix for Storing Syndroms
Errs = eye(N);  % Simulate Error only at a single Unique Bit
for i = 1:N
    S(i,:) = H*Errs(i,:)';
end

disp('Syndromes: ');
disp(S);

%%%%%%%%%%%%%%%%%
% Generate Data %
Data = round(rand(1,B));     % GENERATE RANDOM DATA
X = Data*G;                  % CREATE CODEWORD
C = X;                       % RECEIVED CODEWORD

%%%%%%%%%%%%%%%%%%%%%%%
% Simulate Bit Errors %
if(NumErrors > N)
    error(['Number of simulated errors (',num2str(NumErrors),')',' must be lower than message length(',num2str(N),')!']);
end
for i = 1:NumErrors
    C(i) = mod(C(i) + 1,2); % SIMULATE ERROR
end
 
Syndr = mod(H*C',2);         % CALCULATE SYNDROME
Syndr = transpose(Syndr);    % FIX FOR COMPARISON
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Check for Errors Upon Reception %
% ... Atttempt to Repair
Repaired = 0;
if( sum(Syndr) == 0)
   disp('DATA OK (Zero Syndrome)');
else
    disp('ERROR DETECTED !')

    %%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Try to Repair The Error %
    if(Repair == 1)
        for i = 1:N
            if(isequal(Syndr,S(i,:)))
                disp(horzcat('Repairing Error at index: ',num2str(i)));
                C(i) = mod(C(i)+1,2);
                Repaired = 1;
            end
        end
    end
     
    if(isequal(X,C))
        disp('Repair Success :)');
    end
end
 
%%%%%%%%%%%%%%%%%%%%%%%%%%
% Count amount of Errors %
NumErrorsX = 0;
for i = 1:N
   if(C(i) ~= X(i))
       NumErrorsX = NumErrorsX + 1;
   end
end
 
if(Repaired)
    if(NumErrorsX ~= 0)
        warning(horzcat('Repair failed with a total of: ',num2str(NumErrorsX),' Errors :('));
        if(NumErrorsX > NumErrors )
            warning('Another Error possibly created during a repair-operation!');
        end
    end
else
    warning(horzcat('Repair avoided, leaving a total of: ',num2str(NumErrorsX),' Errors'));
end







